// -*- C++ -*-
/*!
 * @file RtcManager.h
 * @if jp
 * @brief RTComponent ޥ͡㥯饹
 * @else
 * @brief RTComponent manager class
 * @endif
 *
 * @date $Date$
 *
 * Copyright (c) 2003-2004 Noriaki Ando <n-ando@aist.go.jp>
 *          Task-intelligence Research Group,
 *          Intelligent System Research Institute,
 *          National Institute of Industrial Science (AIST),
 *          Japan. All rights reserved.
 *
 * $Id$
 */
#ifndef __RtcManager_h__
#define __RtcManager_h__


// CORBA includes
//#include <RTC.h>

// STL includes
#include <list>
#include <map>
#include <memory>
#include <string>
#include <vector>
#include <fstream>

// ACE includes
#include <ace/DLL.h>

// Boost includes
#include <boost/function.hpp>
#include <boost/bind.hpp>

// CORBA stub include
#include "rtm/idl/RTCManagerSkel.h"
#include "rtm/idl/RTCManagerStub.h"
#include "rtm/RtcModuleProfile.h"
#include "rtm/RtcBase.h"
#include "rtm/RtcFactory.h"

// RTM includes

namespace RTM
{
  
  // Use std:: namespace
  using namespace std;
  
  // Pre-declaration of classes
  
  class RtcManager;
  class RtcNaming;
  class RtcConfig;
  
#define RtcMgrCmd boost::function2<bool, vector<string>&, vector<string>& > 
  
  typedef void (*RtcModuleInitProc)(RtcManager* manager);  
  
  /*!
   * @if jp
   *
   * @class RtcManager
   *
   * @brief RTComponent ޥ͡㥯饹
   *
   * RtcManager ϥݡͥȤΥɡ˴ʤɤΥ饤ե
   * 롣ޤݡͥȤФƳƼ掠ӥ󶡤롣
   *
   * @else
   *
   * @class RtcManager
   *
   * @brief RTComponent manager class.
   *
   * RtcManager manages life cycle of RTComponent for example, loading
   * component library, creating component, deleting component etc.
   * This manager also provide some services for RTComponents.
   *
   * @endif
   */
  class RtcManager :
    public virtual POA_RTM::RTCManager, 
							public virtual PortableServer::RefCountServantBase,
														   public ACE_Task<ACE_MT_SYNCH>
  {
	
  public:
	/*!
	 * @if jp
	 * @brief RtcManager 饹󥹥ȥ饯
	 *
	 * ̾拾ޥɥ饤ȤƤȤ롣
	 *
	 * @else
	 *
	 * @brief A constructor of Rtcmanager class.
	 *
	 * Usually constructor is given command line arguments.
	 *
	 * @endif
	 */
    RtcManager(int argc, char** argv);
	
	/*!
	 * @if jp
	 * @brief RtcManager 饹ǥȥ饯
	 * @else
	 * @brief A destructor of Rtcmanager class.
	 * @endif
	 */    
    virtual ~RtcManager();
	
	/*!
	 * @if jp
	 * @brief RtcManager 饹ǥȥ饯
	 * @else
	 * @brief A destructor of Rtcmanager class.
	 * @endif
	 */    
	//    int open(void *args);
	
	/*!
	 * @if jp
	 * @brief ӥΥåɴؿ
	 * @else
	 * @brief Thread function of service
	 * @endif
	 */    
	int svc(void);
	
    //============================================================
    //
    // <<< CORBA interfaces >>>
    //
    //============================================================
	
	
	/*!
	 * @if jp
	 * @brief [CORBA interface] ⥸塼Υ
	 *
	 * ݡͥȤΥ⥸塼ɤƽؿ¹Ԥ롣
	 *
	 * @param pathname ݡͥȥ⥸塼ե̾
	 * @param initfunc ؿ̾
	 * 
	 * @else
	 * @brief [CORBA interface] Load module
	 *
     * Load shared library of a component by path name, and invoke initialize
	 * function.
	 *
	 * @param pathname Component module file name
	 * @param initfunc Initialize function name
	 *
	 * @endif
	 */  
    virtual RtmRes load(const char* pathname, const char* initfunc);
	
	/*!
	 * @if jp
	 * @brief [CORBA interface] ⥸塼Υ
	 *
	 * ݡͥȤΥ⥸塼򥢥ɤ
	 *
	 * @param pathnae ݡͥȥ⥸塼Υե̾
	 * 
	 * @else
	 * @brief [CORBA interface] Unload module
	 *
     * Unload shared library of a component module.
	 *
	 * @param pathname Component module file name
	 *
	 * @endif
	 */  
    virtual RtmRes unload(const char* pathname);
	
	/*!
	 * @if jp
	 * @brief [CORBA interface] ݡͥȤ
	 *
	 * ݡͥȤΥ󥹥󥹤
	 *
	 * @param comp_name ݡͥȥ⥸塼̾
	 * @param comp_name 󥹥̾ ()
	 * 
	 * @else
	 * @brief [CORBA interface] Create component instance
	 *
     * Create component instance with component module name.
	 *
	 * @param comp_name Component module name
	 * @param comp_name Instance name (return value)
	 *
	 * @endif
	 */  
    virtual RTCBase_ptr create_component(const char* comp_name,
										 CORBA::String_out instance_name);
	
	/*!
	 * @if jp
	 * @brief [CORBA interface] ݡͥȤκ
	 *
	 * ݡͥȤΥ󥹥󥹤
	 *
	 * @param comp_name 󥹥̾
	 * 
	 * @else
	 * @brief [CORBA interface] Delete component instance
	 *
     * Delete component instance with instance name
	 *
	 * @param comp_name Instance name
	 *
	 * @endif
	 */  
    virtual RtmRes delete_component(const char* instance_name);
	
	/*!
	 * @if jp
	 * @brief [CORBA interface] ݡͥ Factory ꥹȤμ
	 *
	 * ݡͥ Factory ΥꥹȤ
	 *
	 * @else
	 * @brief [CORBA interface] Get component factory list
	 *
     * Get component factory list.
	 *
	 * @endif
	 */  
    virtual ComponentFactoryList* component_factory_list();
	
	/*!
	 * @if jp
	 * @brief [CORBA interface] ݡͥȥꥹȤμ
	 *
	 * ݡͥȤΥꥹȤ
	 *
	 * @else
	 * @brief [CORBA interface] Get component list
	 *
     * Get component list.
	 *
	 * @endif
	 */  
    virtual RTCBaseList* component_list();
	
	/*!
	 * @if jp
	 * @brief [CORBA interface] ʰץ󥿥ץ꥿
	 *
	 * ޥ͡δʰץ󥿥ץ꥿ޥɤμ¹
	 *
	 * @else
	 * @brief [CORBA interface] Simple interpreter
	 *
     * Execute simple interpreter.
	 *
	 * @endif
	 */  
    virtual RtmRes command(const char* cmd, CORBA::String_out ret);
	
    // end of CORBA interface definition
    //============================================================
	
	
	
    //============================================================
    //
    // <<< Local interfaces >>>
    //
    //============================================================
	/*!
	 * @if jp
	 * @brief ޥ͡ν
	 *
	 * @else
	 * @brief Initialize manager
	 *
	 * @endif
	 */ 
    void initManager();
	
	/*!
	 * @if jp
	 * @brief ޥ͡μ¹
	 *
	 * @else
	 * @brief Run manager
	 *
	 * @endif
	 */ 
    void runManager();
	
	/*!
	 * @if jp
	 * @brief ޥ͡μ¹(֥å⡼)
	 *
	 * @else
	 * @brief Run manager (No blocking mode)
	 *
	 * @endif
	 */ 
    void runManagerNoBlocking();
	
	/*!
	 * @if jp
	 * @brief ޥ͡㥵ХȤΥƥֲ
	 *
	 * @else
	 * @brief Activate manager servant
	 *
	 * @endif
	 */ 
    bool activateManager();
	
	/*!
	 * @if jp
	 * @brief ⥸塼롼μ¹
	 *
	 * @else
	 * @brief Invoke module initialize routine
	 *
	 * @endif
	 */ 
	void initModuleProc(RtcModuleInitProc proc);
	
	/*!
	 * @if jp
	 * @brief ʰץ󥿥ץ꥿ޥɤϿ
	 *
	 * @else
	 * @brief Register simple interpreter command function
	 *
	 * @endif
	 */ 
    bool createCommand(string cmd,
					   boost::function2<bool,
					   vector<string>&, vector<string>&>  func);
	
	/*!
	 * @if jp
	 * @brief ݡͥȥեȥϿ
	 *
	 * @else
	 * @brief Register component factory
	 *
	 * @endif
	 */ 
    bool registerComponent(RtcModuleProfile& profile,
						   RtcNewFunc new_func,
						   RtcDeleteFunc delete_func);
	
	/*!
	 * @if jp
	 * @brief ݡͥȥեȥϿ
	 *
	 * @else
	 * @brief Register component factory
	 *
	 * @endif
	 */ 
	bool registerComponent(RtcFactoryBase* factory);
	
	
	/*!
	 * @if jp
	 * @brief ݡͥ
	 *
	 * @else
	 * @brief Create component
	 *
	 * @endif
	 */ 
    RtcBase* createComponent(const string& module_name, string& comp_name);
	
	/*!
	 * @if jp
	 * @brief ݡͥ
	 *
	 * @else
	 * @brief Create component
	 *
	 * @endif
	 */ 
    RtcBase* createComponent(const string& module_name);
	
	
	/*!
	 * @if jp
	 * @brief ݡͥȸ
	 *
	 * @else
	 * @brief find component
	 *
	 * @endif
	 */ 
    std::vector<RTCBase_ptr> findComponents(const string& comp_name);
	
	/*!
	 * @if jp
	 * @brief InPort  OutPort ³
	 *
	 * @else
	 * @brief Bind InPort with OutPort
	 *
	 * @endif
	 */ 
    std::string bindInOut(RTCBase_ptr comp_in,
						  const std::string& inp_name,
						  RTCBase_ptr comp_out,
						  const std::string& outp_name,
						  RTM::SubscriptionType sub_type = RTM::OPS_NEW);
	/*!
	 * @if jp
	 * @brief InPort  OutPort ̾ˤ³
	 *
	 * @else
	 * @brief Bind InPort with OutPort by name
	 *
	 * @endif
	 */ 
    std::string bindInOutByName(const std::string& comp_name_in,
								const std::string& inp_name,
								const std::string& comp_name_out,
								const std::string& outp_name,
								RTM::SubscriptionType sub_type = RTM::OPS_NEW);
	
	/*!
	 * @if jp
	 * @brief ORB Υݥ󥿤
	 *
	 * @else
	 * @brief Get pointer to ORB
	 *
	 * @endif
	 */ 
    CORBA::ORB_var getORB();
	
	/*!
	 * @if jp
	 * @brief POA Υݥ󥿤
	 *
	 * @else
	 * @brief Get pointer to POA
	 *
	 * @endif
	 */ 
    PortableServer::POA_var getPOA();
	
    //============================================================
    // Commands
	
    bool loadCmd(const vector<string>& cmd, vector<string>& retval);
    bool unloadCmd(const vector<string>& cmd, vector<string>& retval);
    bool createComponentCmd(const vector<string>& cmd, vector<string>& retval);
    bool listComponent(const vector<string>& cmd, vector<string>& retval);
    bool listModule(const vector<string>& cmd, vector<string>& retval);
    bool commandListCmd(const vector<string>& cmd, vector<string>& retval);

	/*!
	 * @if jp
	 * @brief Log˽񤭹ࡣåա郎졢
	 * ͽꤵ줿logե˽񤭹ޤ롣
	 *
	 * @param loglevel ٥롣(OFF, FATAL, ERROR, WARN, INFO, DEBUG, ALL)
	 * @param msg 񤭹å
	 *
	 * @else
	 * @brief Write to log file. Data and time is inserted before message,
	 * and message will be written to log file.
	 *
	 * @param msg Message 
	 *
	 * @endif
	 */ 
	bool writeLog(int loglevel, std::string msg);

	/*!
	 * @if jp
	 * @brief 顼Log˽񤭹ࡣåա郎졢
	 * ͽꤵ줿logե˽񤭹ޤ롣
	 *
	 * @param loglevel ٥롣(OFF, FATAL, ERROR, WARN, INFO, DEBUG, ALL)
	 * @param msg 񤭹å
	 *
	 * @else
	 * @brief Write to error log file. Data and time is inserted before message,
	 * and message will be written to log file.
	 *
	 * @param msg Message 
	 *
	 * @endif
	 */ 
	bool writeErrLog(int loglevel, std::string msg);

  private:
    /*!
     * Initialize ORB and store ORB's reference
     */
    bool initORB();
	
    /*!
     * Initialize POA and POA manager
     */
    bool initPOA();

	bool initLogger();

	
    /*!
     * Find library path
     */
    bool findPath(string path, string &full_path);
	
    /*!
     * Make initialize function name from libname
     */
    bool libnameToInitFunc(const char* libname,
						   const char* initfunc,
						   string &initfunc_name);
	
	
  private:
	/*!
	 * @if jp
	 * @brief RtcConfig ؤΥݥ
	 * @else
	 * @brief Pointer to RtcConfig
	 * @endif
	 */
    auto_ptr<RtcConfig> m_apConfig;
	
	/*!
	 * @if jp
	 * @brief RtcNaming ؤΥݥ
	 * @else
	 * @brief Pointer to RtcNaming
	 * @endif
	 */
    auto_ptr<RtcNaming> m_apNaming;
	
	/*!
	 * @if jp
	 * @brief ORB ؤΥݥ
	 * @else
	 * @brief Pointer to ORB
	 * @endif
	 */
    CORBA::ORB_var m_pORB;
	
	/*!
	 * @if jp
	 * @brief POA ؤΥݥ
	 * @else
	 * @brief Pointer to POA
	 * @endif
	 */
    PortableServer::POA_var m_pPOA;
	
	/*!
	 * @if jp
	 * @brief POAManager ؤΥݥ
	 * @else
	 * @brief Pointer to POAManager
	 * @endif
	 */
    PortableServer::POAManager_var m_pPOAManager;
	
  	/*!
	 * @if jp
	 * @brief ʰץ󥿥ץ꥿Υޥɥޥå
	 * @else
	 * @brief Command map for simple interpreter
	 * @endif
	 */
    map<string, boost::function2<bool, vector<string>&, vector<string>&> > m_CmdMap;
	
  	/*!
	 * @if jp
	 * @brief ݡͥȥ⥸塼ɥѥ
	 * @else
	 * @brief Component module load path
	 * @endif
	 */
    list<string> m_LibraryLoadPath;
	
    //! Component factory and destructor map
	/*    struct Factory {
		  
	RtcModuleProfile _profile;
	RtcNewFunc _new;
	RtcDeleteFunc _delete;
	int _num;
    };
	*/
	
 	/*!
	 * @if jp
	 * @brief ݡͥȥեȥޥå׹¤
	 * @else
	 * @brief Component factory map
	 * @endif
	 */
	struct FactoryMap {
	  //! Mutex
      ACE_Thread_Mutex _mutex;
	  //! Component name <=> Factory
	  map<string, RtcFactoryBase*> _map;
	};
	
 	/*!
	 * @if jp
	 * @brief ݡͥȥեȥޥå
	 * @else
	 * @brief Component factory map
	 * @endif
	 */
	FactoryMap m_FactoryMap;
	
 	/*!
	 * @if jp
	 * @brief ݡͥȥ⥸塼빽¤
	 * @else
	 * @brief Component module structre
	 * @endif
	 */
    struct DLL {
      int _refCount;
      ACE_DLL _dll;
    };
	
 	/*!
	 * @if jp
	 * @brief ݡͥȥ⥸塼ޥå
	 * @else
	 * @brief Component module map
	 * @endif
	 */
    map<string, DLL> m_Lib; // Library name <=> DLL manager
	struct OstreamMtx
	{
	  ACE_Thread_Mutex _mutex;
	  std::ofstream _stream;
	};
	OstreamMtx m_Log;
	OstreamMtx m_ErrLog;

	
  public:
 	/*!
	 * @if jp
	 * @brief ݡͥȥ󥹥󥹥ǡ١¤
	 * @else
	 * @brief Component instance database
	 * @endif
	 */
    struct ComponentDB {
      RtcBase& _component;
      RtcModuleProfile& _mod_profile;
    };
 	/*!
	 * @if jp
	 * @brief ݡͥȥ󥹥󥹥ǡ١ޥå
	 * @else
	 * @brief Component instance database map
	 * @endif
	 */
    map<string, RtcBase*> m_Components;
	
 	/*!
	 * @if jp
	 * @brief ݡͥȥ⥸塼ؿ
	 * @else
	 * @brief Component module initialization function
	 * @endif
	 */
    typedef bool (*RtcComponentInit)(RtcManager* manager);
	
 	/*!
	 * @if jp
	 * @brief ޥ̾͡
	 * @else
	 * @brief Manager name
	 * @endif
	 */
    string m_ManagerName;
	
  };
  
  
}; // end of namespace RTM

#endif // 
