/*! -*- C++ -*-
 *
 * @file RtcRingBuffer.h
 * @if jp
 * @brief Ring buffer ƥץ졼ȥ饹
 * @else
 * @brief Ring buffer template class
 * @endif
 *
 * Copyright (c) 2003,2004 Noriaki Ando <n-ando@aist.go.jp>
 *          Task-intelligence Research Group,
 *          Intelligent System Research Institute,
 *          National Institute of Industrial Science (AIST), Japan
 *          All rights reserved.
 *
 * $Id$
 */

#ifndef __RtcRingBuffer_h__
#define __RtcRingBuffer_h__

#include <vector>
#include <iostream>
#include <ace/Synch.h>

namespace RTM {
  using namespace std;

  /*!
   * @if jp
   *
   * @class RingBuffer
   *
   * @brief RingBuffer ƥץ졼ȥ饹
   * 
   * ǤդηФƥ󥰥Хåե󶡤 RingBuffer ƥץ졼ȥ饹
   * ǤդĹΥХåեĹĤȤ롣
   * ǿ͡ǸͤμФǤդξѿμФԤ롣
   *
   * @else
   *
   * @class RingBuffer
   *
   * @brief RingBuffer template class
   *
   * RingBuffer template class which provides for any types variable.
   * This ring buffer is able to have any length of buffer.
   * You can get newest and oldest value, value in any position of buffer.
   *
   * @endif
   */

  template <class T> class RingBuffer
  {
  public:
	/*!
	 * @if jp
	 *
	 * @brief RingBuffer 饹󥹥ȥ饯
	 *
	 * RingBuffer<T> 饹Υ󥹥ȥ饯
	 * ѥ᡼ȤͿ T ѿΥХåե롣
	 *
	 * @param length ХåեĹ
	 *
	 * @else
	 *
	 * @brief RingBuffer constructor.
	 *
	 * RingBuffer<T> class constructor.
	 * This constructor create the buffer with "length".
	 *
	 * @param length Buffer length
	 *
	 * @endif
	 */
	RingBuffer(int length)
	  : m_Cond(m_Mutex), m_Oldest(0), m_Newest(length)
	{
	  if (length > 1)
		{
		  m_Length = length;
		}
	  else
		{
		  m_Length = 2;
		}
	  m_Buffer.reserve(m_Length);
	}
	
	/*!
	 * @if jp
	 *
	 * @brief RingBuffer 饹󥹥ȥ饯
	 *
	 * RingBuffer<T> 饹Υ󥹥ȥ饯
	 * ѥ᡼ȤͿ T ѿΥХåե롣
	 *
	 * @param length ХåեĹ
	 * @param inival 
	 *
	 * @else
	 *
	 * @brief RingBuffer constructor.
	 *
	 * RingBuffer<T> class constructor.
	 * This constructor create the buffer with "length".
	 *
	 * @param length Buffer length
	 * @param inival Initial value
	 *
	 * @endif
	 */
	RingBuffer(int length, T inival)
	  : m_Cond(m_Mutex), m_Oldest(0), m_Newest(length)
	{
	  if (length > 1)
		{
		  m_Length = length;
		}
	  else
		{
		  m_Length = 2;
		}
	  
	  m_Buffer.reserve(m_Length);
	  
	  for (int i = 0; i < m_Length; i++)
		{
		  m_Buffer[i] = inival;
		}
	}
	
	/*!
	 * @if jp
	 *
	 * @brief RingBuffer ͤǼ
	 *
	 * RingBuffer ˰ͤǼ롣
	 *
	 * @param value Ǽ T 
	 *
	 * @else
	 *
	 * @brief Store given value into the RingBuffer
	 *
	 * Store given argument value into the RingBuffer.
	 *
	 * @param value Stored value
	 *
	 * @endif
	 */
	virtual void put(const T value)
	{
	  m_Buffer[m_Oldest] = value;
	  
	  ACE_Guard<ACE_Thread_Mutex> guard(m_Mutex);
	  m_Newest = m_Oldest;
	  m_Oldest = (++m_Oldest) % m_Length;
	}

	/*!
	 * @if jp
	 *
	 * @brief ǿͤ
	 *
	 * RingBuffer κǿͤ롣
	 *
	 * @else
	 *
	 * @brief Get the newset value
	 *
	 * Get the newest value stored in the RingBuffer.
	 *
	 * @endif
	 */
	virtual const T& get_new()
	{
	  ACE_Read_Guard<ACE_Thread_Mutex> guard(m_Mutex);

	  return m_Buffer[m_Newest];
	}

	/*!
	 * @if jp
	 *
	 * @brief ǸŤͤ
	 *
	 * RingBuffer κǸŤͤ롣
	 *
	 * @else
	 *
	 * @brief Get the oldest value
	 *
	 * Get the oldest value stored in the RingBuffer.
	 *
	 * @endif
	 */
	virtual T& get_old()
	{
	  ACE_Read_Guard<ACE_Thread_Mutex> guard(m_Mutex);

	  return m_Buffer[m_Oldest];
	}

	/*!
	 * @if jp
	 *
	 * @brief ǿͤ pos ĸŤͤ
	 *
	 * RingBuffer κǿͤ pos ĸŤͤ롣
	 *
	 * @else
	 *
	 * @brief Get pos-th older value from the newest value
	 *
	 * Get pos-th older value from the newset value stored in the RingBuffer.
	 *
	 * @endif
	 */
	virtual T& get_back(int pos)
	{
	  ACE_Read_Guard<ACE_Thread_Mutex> guard(m_Mutex);
	  pos = pos % m_Length;
	  
	  return m_Buffer[(m_Length + m_Newest - pos) % m_Length];
	}

	/*!
	 * @if jp
	 *
	 * @brief ǸŤͤ pos Ŀͤ
	 *
	 * RingBuffer κǸŤͤ pos Ŀͤ롣
	 *
	 * @else
	 *
	 * @brief Get pos-th newer value from the oldest value
	 *
	 * Get pos-th newer value from the oldest value stored in the RingBuffer.
	 *
	 * @endif
	 */
	virtual T& get_front(int pos)
	{
	  ACE_Read_Guard<ACE_Thread_Mutex> guard(m_Mutex);
	  pos = pos % m_Length;
	  
	  return m_Buffer[(m_Oldest + pos) % m_Length];
	}

	/*!
	 * @if jp
	 *
	 * @brief ХåեĹ
	 *
	 * ߤΥХåեĶ
	 *
	 * @else
	 *
	 * @brief Get buffer length
	 *
	 * Get buffer length
	 *
	 * @endif
	 */
	virtual int buff_length()
	{
	  return m_Length;
	}

  protected:
	/*!
	 * @if jp
	 * @brief Mutex
	 * @else
	 * @brief Mutex
	 * @endif
	 */
	ACE_Thread_Mutex m_Mutex;

	/*!
	 * @if jp
	 * @brief ѿ
	 * @else
	 * @brief Condition variable
	 * @endif
	 */
	ACE_Condition<ACE_Thread_Mutex> m_Cond;
	/*!
	 * @if jp
	 * @brief ХåեĹ
	 * @else
	 * @brief Buffer length
	 * @endif
	 */
	int m_Length;
	/*!
	 * @if jp
	 * @brief Хåե
	 * @else
	 * @brief Buffer sequence
	 * @endif
	 */
	vector<T> m_Buffer;
	/*!
	 * @if jp
	 * @brief ǿͤΥХåե
	 * @else
	 * @brief Position of newset value stored in
	 * @endif
	 */
	int m_Newest;
	/*!
	 * @if jp
	 * @brief ǸͤΥХåե
	 * @else
	 * @brief Position of oldest value stored in
	 * @endif
	 */
	int m_Oldest;
  };
 
};


#endif // end of __RtcRingBuffer_h__
